/*
 *    ServiceCluster.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
// leave the package name as is for now - should be changed to something better
// cluster? groups?
package org.greenstone.gsdl3.collection;


import org.greenstone.gsdl3.util.*;
import org.greenstone.gsdl3.core.*;
import org.greenstone.gsdl3.service.*;

// java XML classes we're using
import org.w3c.dom.Document; 
import org.w3c.dom.Node; 
import org.w3c.dom.Element; 
import org.w3c.dom.NodeList; 

import java.io.File;
import java.util.HashMap;

/* ServiceCluster - a groups of services that are related in some way
 * Implements ModuleInterface. Contains a list of services provided by the cluster, along with metadata about the cluster itself.
 * a collection is a special type of cluster
 *  @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 *  @version $Revision: 3502 $
 *  @see ModuleInterface
 */
public class ServiceCluster  
    implements ModuleInterface {


    /** base directory for the site that this cluster belongs to*/
    protected String site_home_ = null;

    /** The name of the cluster - for a collection, this is the collection name*/
    protected String cluster_name_ = null;

    /** a reference to the message router */
    protected MessageRouter router_ = null;
    /** The map of services.
     *
     * Maps Services to ServicesImpl objects
     * @see ServicesImpl
     *
     */
    protected HashMap service_map_=null;

    /** XML converter for String to DOM and vice versa */
    protected XMLConverter converter_=null;

    /** container doc for description elements */
    protected Document doc_ = null;
    /** list of services */
    protected Element service_info_ = null;
    /** list of Metadata */
    protected Element meta_info_ = null;
    /** full description */
    protected Element description_ = null;

    public void setSiteHome(String home) {
	site_home_ = home;
    }

    public void setClusterName(String name) {
	cluster_name_ = name;	
	description_.setAttribute("name", name);
    }  

    public void setMessageRouter(MessageRouter m) {
	router_ = m;
    }

    public ServiceCluster() {
	service_map_ = new HashMap();
	converter_ = new XMLConverter();
	doc_ = converter_.newDOM();
	description_ = doc_.createElement("serviceCluster");
	service_info_ = doc_.createElement("serviceList");
	meta_info_ = doc_.createElement("metadataList");
	
    }

    /**
     * Configures the cluster.
     *
     * gsdlHome and clusterName must be set before configure is called.
     *
     * reads the site configuration file, and configures itself
     * this calls configure(Element) with the XML element node from the config
     * file.
     * configure(Element) should be used if the config file has already been 
     * parsed.  This method will work with any subclass.
     * 
     * @return true if configure successful, false otherwise.
     */
    public boolean configure() {

	if (site_home_ == null || cluster_name_== null) {
	    System.err.println("site_home_ and cluster_name_ must be set before configure called!");
	    return false;
	}
	System.out.println("configuring service cluster");
	// read the site configuration file
	File config_file_ = new File(GSFile.siteConfigFile(site_home_));

	if (!config_file_.exists()) {
	    System.err.println(config_file_+" does not exist");
	    System.err.println("couldn't configure cluster: "+cluster_name_);
	    return false;
	}

	Document doc = converter_.getDOM(config_file_);
	
	// get the appropriate service cluster element
	Element cluster_list = (Element)GSXML.getChildByTagName(doc.getDocumentElement(), "ServiceClusterList");
	Element sc = GSXML.getNamedElement(cluster_list, "ServiceCluster",
					   "name", cluster_name_);
	
	return this.configure(sc);
    }
    
    public boolean configure(Element cluster_info) {
	
	// create all the services
	NodeList nodes = cluster_info.getElementsByTagName("servicesImpl");
	
	// the xml request to send to each servicesImpl to query what
	// services it provides
	Element request = doc_.createElement("request");
	request.setAttribute("type", "describe");
	request.setAttribute("info", "serviceList");

	if (nodes.getLength()==0) {
	    System.err.println("Cluster configuration error: cluster "+cluster_name_+" has no service modules!");
	    return false;
	}

	for(int i=0; i<nodes.getLength(); i++) {
	    Element n = (Element)nodes.item(i);
	    String servicetype = n.getAttribute("name");
	    
	    try {
		ServicesImpl s = (ServicesImpl)Class.forName("org.greenstone.gsdl3.service."+servicetype).newInstance();

		s.setSiteHome(site_home_);
		s.setClusterName(cluster_name_);
		s.setMessageRouter(router_);
		// pass the xml node to the service for configuration
		s.configure(n);
		
		// find out the supported service types for this service module
		Node types = s.process(request);
		NodeList typenodes = ((Element)types).getElementsByTagName("service");	   
		
		for (int j=0; j<typenodes.getLength();j++) {
		    String service = ((Element) typenodes.item(j)).getAttribute("name");	    
		    service_map_.put(service, s);
		    
		    // also add info to the ServiceInfo XML element
		    service_info_.appendChild(doc_.importNode(typenodes.item(j), true));
		}
	    } catch (Exception e) {
		System.out.println("Cluster configure exception: couldn't create service module:org.greenstone.gsdl3.service."+servicetype+"\n"+e.getMessage() + e.getClass() );
		e.printStackTrace();
		//return false;
	    }
	    
	}
	
	// get the metadata - for now just add it to the list
	Element meta_list = (Element) GSXML.getChildByTagName(cluster_info, "metadataList");
	NodeList metanodes = meta_list.getElementsByTagName("metadata");
	
	if (metanodes.getLength()>0) { 	
	    for(int k=0; k<metanodes.getLength(); k++) {
		meta_info_.appendChild(doc_.importNode(metanodes.item(k),true));
	    }
	}	     
    
	// add elements to description
	description_.appendChild(service_info_);
	description_.appendChild(meta_info_);
	return true;
	
	
    }

    /**
     * Process an XML document - uses Strings 
     *  just calls process(Node).
     *
     * @param in the Document to process - a string
     * @return the resultant document as a string - contains any error messages
     * @see String
     */
    public String process(String in) {

	Document doc = converter_.getDOM(in);
	Element e = doc.getDocumentElement();
	
	Element res = process(e);
	return converter_.getString(res);
	
    }
    
    /** process XML as Node
     *
     */
    public Element process(Element message) {

	String to = message.getAttribute("to");

	// the cluster name should be first, check, then remove
	String clustername = GSPath.getFirstLink(to);
	if (clustername.equals(cluster_name_)){
	    to = GSPath.removeFirstLink(to);
	    message.setAttribute("to", to);
	} else {
	    System.out.println("ServiceCluster error: cluster name wrong! was "+clustername+" should have been "+cluster_name_);
	    return null;
	}

	if (to.equals("")) { // this command is for me
	    String type = message.getAttribute("type");
	    if (type.equals("describe")) {
		// create a response with the description inside it
		Element response = doc_.createElement("response");
		response.setAttribute("from", cluster_name_);
		response.setAttribute("type", "describe");
		response.appendChild(description_);
		return response;
	    } else {
		System.out.println("ServiceCluster: error, cant handle request:");
		System.out.println(converter_.getString(message));
		return null;
	    }
	    
	} else { // the request is for one of my services
	    String service = GSPath.getFirstLink(to);
	    
	    if (service_map_.containsKey(service)) {
		return ((ModuleInterface)service_map_.get(service)).process(message);
	    } else {
		System.err.println("ServiceCluster: non-existant service, "+service+", specified, in:\n"+converter_.getString(message));
		return null;
	    }

	}
    }
	
}






